require 'logger'
require 'bot_mob/core_ext/hash'
require 'bot_mob/core_ext/string'
require 'bot_mob/version'

# # BotMob
#
# BotMob base module namespace
module BotMob
  class InvalidNetworkError < StandardError; end
  class UnspecifiedTargetNetworkError < StandardError; end
  class UnconnectedTargetNetworkError < StandardError; end
  class UndefinedResponseError < StandardError; end
  class NetworkConnectionError < StandardError; end

  autoload :Bot,              'bot_mob/bot'
  autoload :Environment,      'bot_mob/environment'
  autoload :Command,          'bot_mob/command'
  autoload :Connection,       'bot_mob/connection'
  autoload :InboundMessage,   'bot_mob/inbound_message'
  autoload :Logger,           'bot_mob/logger'
  autoload :OutboundMessage,  'bot_mob/outbound_message'
  autoload :Roster,           'bot_mob/roster'
  autoload :Server,           'bot_mob/server'

  # Networks
  require 'bot_mob/networks/roaming'
  require 'bot_mob/networks/slack'

  # ## `start!`
  #
  # Start the bot mob server
  #
  # Usage:
  #
  #    BotMob.start! do |app|
  #      app.register MyBot
  #    end
  def self.start!
    roster = BotMob::Roster.new
    yield(roster)
    BotMob::Server.run! roster: roster
  end

  # ## `root`
  #
  # Provide the base root path of the gem directory
  def self.root
    File.expand_path('../..', __FILE__)
  end

  # ## `env`
  #
  # The specified environment for your bot's runtime
  def self.env
    @environment ||= BotMob::Environment.new(ENV['MOB_ENV'])
  end

  # ## `logger`
  #
  # Provide a logger for the BotMob gem, silences test env output by default
  def self.logger
    @logger ||= begin
      buffer = BotMob::Logger.new(ENV['MOB_ENV'] == 'test' ? '/dev/null' : STDOUT)
      if buffer
        buffer.level = ($PROGRAM_NAME == 'irb' ? Logger::DEBUG : Logger::INFO)
        buffer.formatter = proc { |_s, _d, _p, msg| "#{msg}\n" }
      end
      buffer
    end
  end
end
